"""
These tests verify that the crema api works as intended:
1. dataset objects are properly created
2. fdr methods are properly executed
3. result objects are properly created
"""

import pytest
from crema.parsers import *
from crema.methods import *


# The following dataframes are what I use as "actual" in the following tests:
# i.e. this is what the pandas dataframe should look like after the dataset/result object is created

# Used in test_single_basic_data, test_single_int_targets_data
testframe_single_basic_data = pd.DataFrame(
    {
        "scan": [1, 2, 3, 4, 5, 1, 2, 3, 4, 5],
        "combined p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.25,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
        ],
        "target/decoy": [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            False,
        ],
    }
)

# Used in test_single_basic_tdc, test_single_int_targets_tdc
testframe_single_basic_tdc = pd.DataFrame(
    {
        "scan": [3, 2, 5, 4, 1],
        "combined p-value": [0.1, 0.2, 0.25, 0.55, 0.6],
        "target/decoy": [True, False, True, True, True],
        "FDR": [1, 1, 1, 2 / 3, 1 / 2],
        "Q_Value": [0.5, 0.5, 0.5, 0.5, 0.5],
    }
)

# Used in test_single_text_scan_data
testframe_single_text_scan_data = pd.DataFrame(
    {
        "scan": [
            "C:/test/string/fake/file/path/1",
            "C:/test/string/fake/file/path/2",
            "C:/test/string/fake/file/path/3",
            "C:/test/string/fake/file/path/4",
            "C:/test/string/fake/file/path/5",
            "C:/test/string/fake/file/path/1",
            "C:/test/string/fake/file/path/2",
            "C:/test/string/fake/file/path/3",
            "C:/test/string/fake/file/path/4",
            "C:/test/string/fake/file/path/5",
        ],
        "combined p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.25,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
        ],
        "target/decoy": [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            False,
        ],
    }
)

# Used in test_single_text_scan_tdc
testframe_single_text_scan_tdc = pd.DataFrame(
    {
        "scan": [
            "C:/test/string/fake/file/path/3",
            "C:/test/string/fake/file/path/2",
            "C:/test/string/fake/file/path/5",
            "C:/test/string/fake/file/path/4",
            "C:/test/string/fake/file/path/1",
        ],
        "combined p-value": [0.1, 0.2, 0.25, 0.55, 0.6],
        "target/decoy": [True, False, True, True, True],
        "FDR": [1, 1, 1, 2 / 3, 1 / 2],
        "Q_Value": [0.5, 0.5, 0.5, 0.5, 0.5],
    }
)

# Used in test_single_add_spectrum
testframe_single_add_spectrum_data = pd.DataFrame(
    {
        "scan": [1, 2, 3, 4, 5, 1, 2, 3, 4, 5],
        "combined p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.25,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
        ],
        "target/decoy": [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            False,
        ],
        "extras": [
            "a",
            "b",
            "c",
            "d",
            "e",
            "z",
            "b",
            "y",
            "d",
            "x",
        ],
    }
)

# Used in test_single_add_spectrum_tdc
testframe_single_add_spectrum_tdc = pd.DataFrame(
    {
        "scan": [3, 2, 5, 5, 4, 1, 1, 3],
        "combined p-value": [0.1, 0.2, 0.25, 0.3, 0.55, 0.6, 0.7, 0.7],
        "target/decoy": [True, False, True, False, True, True, True, False],
        "extras": ["c", "b", "e", "x", "d", "z", "a", "y"],
        "FDR": [1, 1, 1, 1, 1, 3 / 4, 3 / 5, 4 / 5],
        "Q_Value": [3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5, 3 / 5, 4 / 5],
    }
)

# Used in test_single_arbitrary_data
testframe_single_arbitrary_data = pd.DataFrame(
    {
        "scan": [1, 2, 3, 4, 5, 1, 2, 3, 4, 5],
        "combined p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.3,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
        ],
        "target/decoy": [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            False,
        ],
    }
)

# Used in test_single_arbitrary_tdc
# Result if TDC arbitrarily chooses True
testframe_single_arbitrary_tdc_true = pd.DataFrame(
    {
        "scan": [3, 2, 5, 4, 1],
        "combined p-value": [0.1, 0.2, 0.3, 0.55, 0.6],
        "target/decoy": [True, False, True, True, True],
        "FDR": [1, 1, 1, 2 / 3, 1 / 2],
        "Q_Value": [0.5, 0.5, 0.5, 0.5, 0.5],
    }
)

# Used in test_single_arbitrary_tdc
# Result if TDC arbitrarily chooses False
testframe_single_arbitrary_tdc_false = pd.DataFrame(
    {
        "scan": [3, 2, 5, 4, 1],
        "combined p-value": [0.1, 0.2, 0.3, 0.55, 0.6],
        "target/decoy": [True, False, False, True, True],
        "FDR": [1.0, 1.0, 1.0, 1.0, 1.0],
        "Q_Value": [1.0, 1.0, 1.0, 1.0, 1.0],
    }
)

# Used in test_single_noncrux_data
testframe_single_noncrux_data = pd.DataFrame(
    {
        "scan": [1, 2, 3, 4, 5, 1, 2, 3, 4, 5],
        "p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.25,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
        ],
        "target": [
            True,
            False,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            False,
        ],
    }
)

# Used in test_single_noncrux_tdc
testframe_single_noncrux_tdc = pd.DataFrame(
    {
        "scan": [3, 2, 5, 4, 1],
        "p-value": [0.1, 0.2, 0.25, 0.55, 0.6],
        "target": [True, False, True, True, True],
        "FDR": [1, 1, 1, 2 / 3, 1 / 2],
        "Q_Value": [0.5, 0.5, 0.5, 0.5, 0.5],
    }
)

# Used in test_multi_data
testframe_multi_data = pd.DataFrame(
    {
        "scan": [1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5],
        "combined p-value": [
            0.7,
            0.4,
            0.1,
            0.55,
            0.3,
            0.6,
            0.2,
            0.7,
            0.56,
            0.3,
            0.4,
            0.7,
            0.59,
            0.55,
            0.9,
            0.44,
            0.75,
            0.6,
            0.5,
            0.89,
        ],
        "target/decoy": [
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ],
    }
)

# Used in test_multi_tdc
testframe_multi_tdc = pd.DataFrame(
    {
        "scan": [3, 2, 5, 1, 4],
        "combined p-value": [0.1, 0.2, 0.3, 0.4, 0.5],
        "target/decoy": [True, True, True, False, False],
        "FDR": [1, 1 / 2, 1 / 3, 2 / 3, 1],
        "Q_Value": [1 / 3, 1 / 3, 1 / 3, 2 / 3, 1],
    }
)


@pytest.fixture
def test_single_basic_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_basic.csv" to use in
    subsequent test cases. This file has crux
    default column names.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(["data/single_basic.csv"])
    return psm


@pytest.fixture
def test_single_int_targets_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_int_targets.csv" to use in
    subsequent test cases. This file has crux default
    column names. However, note that target column
    contains integer values (0 and 1), but also
    works with (-1 and 1).

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(["data/single_int_targets.csv"])
    return psm


@pytest.fixture
def test_single_text_scan_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_text_scan.csv" to use in
    subsequent test cases. This file has crux default
    column names. However, note that scan column
    contains String identifiers.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(["data/single_text_scan.csv"])
    return psm


@pytest.fixture
def test_single_add_spectrum_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_add_spectrum.csv" to use in
    subsequent test cases. This file has crux default
    column names with an additional "extras" column
    to serve as an additional spectrum column input.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(
        ["data/single_add_spectrum.csv"], spectrum_col=["scan", "extras"]
    )
    return psm


@pytest.fixture
def test_single_arbitrary_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_arbitrary.csv" to use in
    subsequent test cases. This file has crux default
    column names. However, there are target/decoy psms
    with equal p-value to force an arbitrary choice.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(["data/single_arbitrary.csv"])
    return psm


@pytest.fixture
def test_single_noncrux_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "single_noncrux.csv" to use in
    subsequent test cases. This file has non-crux
    column names:
    "scan", "p-value", "target".

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    psm = read_file(["data/single_noncrux.csv"], "scan", "p-value", "target")
    return psm


@pytest.fixture
def test_multi_dataset_class():
    """
    Creates a pytest fixture of a PsmDataset object
    by reading in "multi_target.csv" and
    "multi_target.csv" to use in subsequent test cases.
    Both of these files have crux default
    column names.

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited files.
    """
    files = ["data/multi_target.csv", "data/multi_decoy.csv"]
    psm = read_file(files)
    return psm


def test_single_basic_data(test_single_basic_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file in its most basic
    form with crux default column names. The point of this test is
    to ensure that there are no errors with the most basic input.

    Parameters
    ----------
    test_single_basic_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_basic.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_basic_data"
    """
    actual = testframe_single_basic_data.copy()
    compare = test_single_basic_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_basic_tdc(test_single_basic_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_basic_data.

    Parameters
    ----------
    test_single_basic_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_basic.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_single_basic_tdc"
    """
    actual = testframe_single_basic_tdc.copy()
    output = calculate_tdc(test_single_basic_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_int_targets_data(test_single_int_targets_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file with crux default
    column names and target column values containing integers.
    The point of this test is to make sure calculations can be performed
    with integer identifiers (1/0 or 1/-1) for target/decoy psms.

    Parameters
    ----------
    test_single_int_targets_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_int_targets.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_basic_data"
    """
    actual = testframe_single_basic_data.copy()
    compare = test_single_int_targets_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_int_targets_tdc(test_single_int_targets_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_int_targets_data.

    Parameters
    ----------
    test_single_int_targets_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_int_targets.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_basic_tdc"
    """
    actual = testframe_single_basic_tdc.copy()
    output = calculate_tdc(test_single_int_targets_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_text_scan_data(test_single_text_scan_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file with crux default
    column names and scan column values containing Strings.
    The point of this test is to make sure calculations can be performed
    with non integer scan identifiers, i.e. filename/#.

    Parameters
    ----------
    test_single_text_scan_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_text_scan.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_text_scan_data"
    """
    actual = testframe_single_text_scan_data.copy()
    compare = test_single_text_scan_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_text_scan_tdc(test_single_text_scan_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_text_scan_data.

    Parameters
    ----------
    test_single_text_scan_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_text_scan.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_single_text_scan_tdc"
    """
    actual = testframe_single_text_scan_tdc.copy()
    output = calculate_tdc(test_single_text_scan_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_add_spectrum_data(test_single_add_spectrum_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file with crux default
    column names and an additional spectrum column.
    The point of this test is to make sure the user can specify more
    than one spectrum columns if the need arises.

    Parameters
    ----------
    test_single_add_spectrum_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_add_spectrum.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_add_spectrum_data"
    """
    actual = testframe_single_add_spectrum_data.copy()
    compare = test_single_add_spectrum_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_add_spectrum_tdc(test_single_add_spectrum_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_add_spectrum_data.

    Parameters
    ----------
    test_single_add_spectrum_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_add_spectrum.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_single_add_spectrum_tdc"
    """
    actual = testframe_single_add_spectrum_tdc.copy()
    output = calculate_tdc(test_single_add_spectrum_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_arbitrary_data(test_single_arbitrary_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file with crux default
    column names and duplicate target/decoy psms of equal p-value.
    The point of this test is to make sure that crema arbitrarily
    chooses which psm to keep.

    Parameters
    ----------
    test_single_arbitrary_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_arbitrary.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_arbitrary_datadata"
    """
    actual = testframe_single_arbitrary_data.copy()
    compare = test_single_arbitrary_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_arbitrary_tdc(test_single_arbitrary_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_arbitrary_data.

    Parameters
    ----------
    test_single_arbitrary_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_arbitrary.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_single_arbitrary_tdc"
    """
    actual_true = testframe_single_arbitrary_tdc_true.copy()
    actual_false = testframe_single_arbitrary_tdc_false.copy()
    output = calculate_tdc(test_single_arbitrary_dataset_class)
    compare = output.data
    if compare.iloc[2, 2]:
        # If arbitrarily chooses True
        pd.testing.assert_frame_equal(actual_true, compare)
    else:
        # If arbitrarily chooses False
        pd.testing.assert_frame_equal(actual_false, compare)


def test_single_noncrux_data(test_single_noncrux_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in a single tab delimited file with non crux
    column names. The point of this test is to make sure the user
    can supply an input file with differing column names by specifying
    the names of the new column names.

    Parameters
    ----------
    test_single_noncrux_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_noncrux.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_single_noncrux_data"
    """
    actual = testframe_single_noncrux_data.copy()
    compare = test_single_noncrux_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_single_noncrux_tdc(test_single_noncrux_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_single_noncrux_data.

    Parameters
    ----------
    test_single_noncrux_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "single_noncrux.csv" file

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_single_noncrux_tdc"
    """
    actual = testframe_single_noncrux_tdc.copy()
    output = calculate_tdc(test_single_noncrux_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)


def test_multi_data(test_multi_dataset_class):
    """
    Checks whether or not a PsmDataset object is created properly
    after reading in multiple (two) tab delimited files with crux default
    column names. The point of this test is to ensure that multiple files
    can be combined to perform analysis.

    Parameters
    ----------
    test_multi_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "multi_target.csv" file
        and the "multi_decoy.csv" file.

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in PsmDataset object
        is equal to the dataframe named "testframe_multi_data"
    """
    actual = testframe_multi_data.copy()
    compare = test_multi_dataset_class.data
    pd.testing.assert_frame_equal(actual, compare)


def test_multi_tdc(test_multi_dataset_class):
    """
    Checks whether or not a Result object is created properly
    after executing the calculate_tdc method on a PsmDataset object
    from test_multi_data.

    Parameters
    ----------
    test_multi_dataset_class : pytest fixture of a PsmDataset object
        A psm object created by reading in the "multi_target.csv" file
        and the "multi_decoy.csv" file.

    Returns
    -------
    Pandas Assert Frame
        Asserts whether or not the data in Result object
        is equal to the dataframe named "testframe_multi_tdc"
    """
    actual = testframe_multi_tdc.copy()
    output = calculate_tdc(test_multi_dataset_class)
    compare = output.data
    pd.testing.assert_frame_equal(actual, compare)



"""
This module contains the parsers for reading in PSMs
"""

import pandas as pd
import numpy as np
from .dataset import PsmDataset


def read_file(
    input_files,
    spectrum_col="scan",
    score_col="combined p-value",
    target_col="target/decoy",
    delimiter="," or "\t",
):
    """
    Read tab-delimited files.

    Parameters
    ----------
    input_files : str or tuple of str
        one or more tab-delimited file(s) to read
    spectrum_col : str or tuple of str
        one or more column names that identify the psm
    score_col : str
        name of the column that defines the scores (p-values) of the psms
    target_col : str
        name of the column that indicates if a psm is a target/decoy
    delimiter : str
        string character equal to what is used to separate columns
        within the tab-delimited file

    Returns
    -------
    PsmDataset
        A :py:class:`~crema.dataset.PsmDataset` object
        containing the PSM data from the given tab-delimited file.
    """
    # Store column names in a list to be used by read_csv method
    fields = []
    if type(spectrum_col) == str:
        fields = [spectrum_col, score_col, target_col]
    else:
        for col in spectrum_col:
            fields.append(col)
        fields.append(score_col)
        fields.append(target_col)
    # fields = [spectrum_col, score_col, target_col]
    # Create empty Pandas dataframe
    data = pd.DataFrame()
    # Loop through all given files
    for file in input_files:
        data = data.append(
            pd.read_csv(file, sep=delimiter, usecols=fields), ignore_index=True
        )
    data = convert_target_col(data, target_col)
    return PsmDataset(data, spectrum_col, score_col, target_col)


def convert_target_col(data, target_col):
    """
    Convert values in target column to boolean True/False.

    Parameters
    ----------
    data : pandas.DataFrame
        A pandas.DataFrame of the data before the target/decoy column has been converted to boolean
    target_col : str
        name of the column that indicates if a psm is a target/decoy

    Returns
    -------
    data : pandas.DataFrame
        A pandas.DataFrame of the data after the target/decoy column has been converted to boolean
    """
    # Grab the first value in the target column
    identifier = data.iloc[0][target_col]
    # If the first value is already a boolean, return the data without manipulating anything
    if isinstance(identifier, bool):
        return data
    # If the first value is a numeric value, convert to boolean
    elif isinstance(identifier, (float, int, np.int64)):
        targets = {
            not (0 and 0.0 and -1 and -1.0): True,
            0 or 0.0: False,
            -1 or -1.0: False,
        }
        data[target_col] = data[target_col].map(targets)
    # If the first value is a string, convert to boolean
    elif isinstance(identifier, str):
        targets = {
            "target": True,
            "t": True,
            "decoy": False,
            "d": False,
            "f": False,
        }
        data[target_col] = data[target_col].map(targets)
    return data



"""
Initialize the crema package.
"""
from pkg_resources import get_distribution, DistributionNotFound


try:
    __version__ = get_distribution("crema-ms").version
except DistributionNotFound:
    pass

# Here is where we can export public functions and classes.
from crema.dataset import PsmDataset
from crema.methods import calculate_tdc
from crema.parsers import *
from crema.result import Result
from crema.params import Params



"""
Setup the crema package.
"""
import setuptools

with open("README.md", "r") as readme:
    LONG_DESC = readme.read()

DESC = (
    "Confidence estimation for peptide detection in mass spectrometry"
    " proteomics"
)

CATAGORIES = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]

setuptools.setup(
    name="crema-ms",
    author="William E. Fondrie",
    author_email="fondriew@gmail.com",
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    url="https://github.com/Noble-Lab/crema",
    packages=setuptools.find_packages(),
    license="Apache 2.0",
    entry_points={"console_scripts": ["crema = crema.crema:main"]},
    classifiers=CATAGORIES,
    install_requires=["numpy>=1.18.1", "pandas>=1.0.3", "numba>=0.48.0"],
    use_scm_version=True,
    setup_requires=["setuptools-scm"],
    extras_require={
        "docs": [
            "numpydoc>=1.0.0",
            "sphinx-argparse>=0.2.5",
            "sphinx-rtd-theme>=0.5.0",
            "nbsphinx>=0.7.1",
            "ipykernel>=5.3.0",
        ],
        "dev": ["pre-commit>=2.7.1", "black>=19.10b0", "pytest>=6.0.1"],
    },
)



"""
This module contains the various methods for calculating FDRs and q-values
"""

import random
from .result import Result


def calculate_tdc(psm):
    """
    Calculates FDR and Q-Value using Target-Decoy Competition

    Parameters
    ----------
    psm : A :py:class:`~crema.dataset.PsmDataset` object
        dataset to apply FDR calculation upon

    Returns
    -------
    data : pandas.DataFrame
        A pandas.DataFrame of the data from the original dataset with additional FDR and Q-Value columns
    """

    # copy psm dataframe from dataset object - don't want to manipulate the original
    data = psm.data.copy()

    # note down all column names from dataset object
    spectrum_col = psm.spectrum_col
    score_col = psm.score_col
    target_col = psm.target_col

    # determine items to sort by
    sort_order = []
    if type(spectrum_col) == str:
        sort_order = [spectrum_col, score_col, target_col]
    else:
        for col in spectrum_col:
            sort_order.append(col)
        sort_order.append(score_col)
        sort_order.append(target_col)

    # sort dataframe by spectrum and p-value ascending
    data = data.sort_values(by=sort_order)

    # look through and delete all duplicate psms with higher p-values
    data = _delete_duplicates(data, spectrum_col, score_col, target_col)

    # sort dataframe by spectrum and p-value ascending
    data = data.sort_values(by=[score_col])

    # calculate fdr at each psm
    _calculate_fdr(data, target_col)

    # calculate q value at each psm
    _calculate_q_value(data, "FDR")

    # return a result object containing the manipulated data and respective calculations
    return Result(data, spectrum_col, score_col, target_col)


def _compare_spectrum_col(spectrum_col, curr_row, next_row):
    if type(spectrum_col) == str:
        return curr_row[spectrum_col] != next_row[spectrum_col]
    else:
        for col in spectrum_col:
            if curr_row[col] != next_row[col]:
                return True
        return False


def _delete_duplicates(data, spectrum_col, score_col, target_col):
    """
    Deletes duplicate PSMs based on unique spectrum identifier

    Parameters
    ----------
    data : pandas.DataFrame
        dataframe of PSMs of data from original dataset object (MUST BE SORTED BY SPECTRUM, SCORE, TARGET)
    spectrum_col : str
        name of the column that identifies the psm
    score_col : str
        name of the column that defines the scores (p-values) of the psms
    target_col : str
        name of the column that indicates if a psm is a target/decoy

    Returns
    -------
    data : pandas.DataFrame
        A pandas.DataFrame of the data from the original dataset with additional FDR and Q-Value columns
    """

    prev_score = 1
    prev_target = True
    prev_index = data.iloc[-1].name
    for index, row in data.iterrows():
        # If spectrum col in current row is different from next row, increment curr variables and move on
        if _compare_spectrum_col(
            spectrum_col, data.loc[prev_index], data.loc[index]
        ):
            prev_score = row[score_col]
            prev_target = row[target_col]
            prev_index = index
        # If they are the same, check if the score/target values are the same and remove accordingly
        else:
            if row[score_col] == prev_score:
                if row[target_col] == prev_target:
                    data = data.drop(index)
                else:
                    prev_target = not prev_target
                    decision = random.randint(1, 2)
                    if decision == 1:
                        data.loc[prev_index, target_col] = True
                    else:
                        data.loc[index, target_col] = False
                    data = data.drop(index)
            else:
                data = data.drop(index)
    return data


def _calculate_fdr(data, target_col):
    """
    Calculates FDR at each PSM

    Parameters
    ----------
    data : pandas.DataFrame
        dataframe of PSMs of data from original dataset object (MUST BE SORTED BY SCORE ASCENDING)
    target_col : str
        name of the column that indicates if a psm is a target/decoy
    """
    # List to store fdr at each psm
    fdr = []
    # Keep track of # of targets/decoys along the way
    target = 0
    decoy = 0
    for value in data[target_col]:
        if value:
            target += 1
        else:
            decoy += 1
        fdr.append(min((decoy + 1) / target, 1))
    data["FDR"] = fdr


def _calculate_q_value(data, fdr_col):
    """
    Calculates Q_Value at each PSM

    Parameters
    ----------
    data : pandas.DataFrame
        dataframe of PSMs of data from original dataset object (MUST BE SORTED BY SCORE ASCENDING)
    fdr_col : str
        name of the column that indicates the FDR of each psm
    """
    # List to store q_value at each psm
    q_val = []
    lowest_q_value = data[fdr_col][data.index[-1]]
    for index, row in data[::-1].iterrows():
        curr_q_value = row[fdr_col]
        if curr_q_value < lowest_q_value:
            lowest_q_value = curr_q_value
        q_val.append(lowest_q_value)
    q_val.reverse()
    data["Q_Value"] = q_val



"""
The :py:class:`Params` class is used to define the details and arguments necessary
for running crema from the command line.
"""

import argparse
import os
import crema


class Params:
    """
    All possible arguments and parameters for crema's CLI.
    Options can be specified as command-line arguments.
    """

    def __init__(self):
        """
        Initialize a Params object that holds an argparse parser.
        """
        self.parser = _configure_parser()


def _configure_parser():
    """Creates and configures all the arguments for the parser"""

    desc = (
        f"crema version {crema.__version__}\n\n"
        "Written by Donavan See (seed99@cs.washington.edu)\n\n"
        "Department of Genome Sciences at the University of Washington\n\n"
        "Official code website: <link tbd>\n\n"
        "More documentation and examples: <link tbd>"
    )

    parser = argparse.ArgumentParser(
        description=desc,
    )

    parser.add_argument(
        "input_files",
        type=str,
        nargs="+",
        help="One or more tab-delimited file(s) to read",
    )

    parser.add_argument(
        "--score",
        type=str,
        default="combined p-value",
        help="name of the column that defines the scores (p-values) of the psms."
        "\n Expects decimal or float column values",
    )

    parser.add_argument(
        "--spectrum",
        type=str,
        nargs="+",
        default="scan",
        help="one or more column names that identify the psms."
        "\n Expects numeric or string column values",
    )

    parser.add_argument(
        "--target",
        type=str,
        default="target/decoy",
        help="name of the column that indicates if a psm is a target/decoy."
        "\n Expects column values containing any of the following combinations: (True/False), (target/decoy),"
        "(t/d), (t/f), (1/0), (1/-1)",
    )

    parser.add_argument(
        "--file_root",
        type=str,
        help="This string will be added as a prefix to all output file names",
    )

    parser.add_argument(
        "--output_dir",
        type=str,
        help="The directory where output files will be created. Defaults to current working directory.",
    )

    return parser
